# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import MagicMock
from allensdk.api.queries.synchronization_api import SynchronizationApi


@pytest.fixture
def synch():
    sa = SynchronizationApi()
    sa.json_msg_query = MagicMock(name='json_msg_query')

    return sa


def test_image_to_image(synch):
    '''
    Notes
    -----
    Expected link is slightly modified for json and float serialization of zeros.

    See: `Image Alignment `<http://help.brain-map.org/display/mouseconnectivity/API#API-ImageAlignment>_
    , link labeled 'Sync a VISp and VISal experiment to a location in a SCs SectionDataSet'.
    '''
    section_image_id = 114754496
    (x, y) = (18232, 10704)
    section_data_set_ids = [113887162, 116903968]

    _ = synch.get_image_to_image(section_image_id,
                                 x, y,
                                 section_data_set_ids)
    expected = 'http://api.brain-map.org/api/v2/image_to_image/114754496.json?x=18232.000000&y=10704.000000&section_data_set_ids=113887162,116903968'
    synch.json_msg_query.assert_called_once_with(expected)


def test_image_to_image_2d(synch):
    section_image_id = 68173101
    (x, y) = (6208, 2368)
    section_image_ids = [68173103, 68173105, 68173107]

    _ = synch.get_image_to_image_2d(section_image_id,
                                    x, y,
                                    section_image_ids)
    expected = 'http://api.brain-map.org/api/v2/image_to_image_2d/68173101.json?x=6208.000000&y=2368.000000&section_image_ids=68173103,68173105,68173107'
    synch.json_msg_query.assert_called_once_with(expected)


def test_reference_to_image(synch):
    reference_space_id = 10
    (x, y, z) = (6085, 3670, 4883)
    section_data_set_ids = [68545324, 67810540]

    _ = synch.get_reference_to_image(reference_space_id,
                                     x, y, z,
                                     section_data_set_ids)
    expected = 'http://api.brain-map.org/api/v2/reference_to_image/10.json?x=6085.000000&y=3670.000000&z=4883.000000&section_data_set_ids=68545324,67810540'
    synch.json_msg_query.assert_called_once_with(expected)


def test_image_to_reference(synch):
    section_image_id = 68173101
    (x, y) = (6208, 2368)

    _ = synch.get_image_to_reference(section_image_id,
                                     x, y)
    expected = 'http://api.brain-map.org/api/v2/image_to_reference/68173101.json?x=6208.000000&y=2368.000000'
    synch.json_msg_query.assert_called_once_with(expected)


def test_structure_to_image(synch):
    section_data_set_id = 68545324
    structure_ids = [315, 698, 1089, 703, 477,
                     803, 512, 549, 1097, 313, 771, 354]

    _ = synch.get_structure_to_image(section_data_set_id,
                                     structure_ids)
    expected = 'http://api.brain-map.org/api/v2/structure_to_image/68545324.json?structure_ids=315,698,1089,703,477,803,512,549,1097,313,771,354'
    synch.json_msg_query.assert_called_once_with(expected)


def test_image_to_atlas(synch):
    '''
    Notes
    -----
    Expected link is slightly modified for json and float serialization of zeros.

    See: `Image Alignment `<http://help.brain-map.org/display/mouseconnectivity/API#API-ImageAlignment>_
    , link labeled 'Sync the P56 coronal reference atlas to a location in the SCs SectionDataSet'.
    '''
    section_image_id = 114754496
    (x, y) = (18232, 10704)
    atlas_id = 1
    _ = synch.get_image_to_atlas(section_image_id,
                                 x, y,
                                 atlas_id)
    expected = 'http://api.brain-map.org/api/v2/image_to_atlas/114754496.json?x=18232.000000&y=10704.000000&atlas_id=1'
    synch.json_msg_query.assert_called_once_with(expected)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from allensdk.api.queries.glif_api import GlifApi
import numpy as np
import pytest
import os

@pytest.fixture
def neuronal_model_id():
    return 566283950

@pytest.fixture
def specimen_id():
    return 325464516

@pytest.fixture
def glif_api():
    endpoint = None

    if 'TEST_API_ENDPOINT' in os.environ:
        endpoint = os.environ['TEST_API_ENDPOINT']
        return GlifApi(endpoint)
    else:
        return None


@pytest.mark.skipif(glif_api() is None, reason='No TEST_API_ENDPOINT set.')
def test_get_neuronal_model_templates(glif_api):

    assert len(glif_api.get_neuronal_model_templates()) == 7

    for template in glif_api.get_neuronal_model_templates():

        if template['id'] == 329230710:
            assert 'perisomatic' in template['name']
        elif template['id'] == 395310498:
            assert '(LIF-R-ASC-A)' in template['name']
        elif template['id'] == 395310469:
            assert '(LIF)' in template['name']
        elif template['id'] == 395310475:
            assert '(LIF-ASC)' in template['name']
        elif template['id'] == 395310479:
            assert '(LIF-R)' in template['name']
        elif template['id'] == 471355161:
            assert '(LIF-R-ASC)' in template['name']
        elif template['id'] == 491455321:
            assert 'Biophysical - all active' in template['name']
        else:
            raise Exception('Unrecognized template: %s (%s)' % (template['id'], template['name']))


@pytest.mark.skipif(glif_api() is None, reason='No TEST_API_ENDPOINT set.')
def test_get_neuronal_models(glif_api, specimen_id):

    model = glif_api.get_neuronal_models([specimen_id])

    assert len(model) == 1
    assert len(model[0]['neuronal_models']) == 2


@pytest.mark.skipif(glif_api() is None, reason='No TEST_API_ENDPOINT set.')
def test_get_neuron_configs(glif_api, specimen_id):
    model = glif_api.get_neuronal_models([specimen_id])

    neuronal_model_id = model[0]['neuronal_models'][0]['id']
    assert neuronal_model_id == 566283950

    np.testing.assert_almost_equal(glif_api.get_neuron_configs([neuronal_model_id])[neuronal_model_id]['th_inf'], 0.024561992461740227)

@pytest.mark.skipif(glif_api() is None, reason='No TEST_API_ENDPOINT set.')
def test_deprecated(glif_api, neuronal_model_id):

    # Exercising deprecated functionality
    len(glif_api.list_neuronal_models())

    glif_api.get_neuronal_model(neuronal_model_id)

    glif_api.get_neuronal_model(neuronal_model_id)
    print(glif_api.get_ephys_sweeps())

    glif_api.get_neuronal_model(neuronal_model_id)
    x = glif_api.get_neuron_config()

    glif_api.get_neuronal_model(neuronal_model_id)
    glif_api.cache_stimulus_file('tmp.nwb')



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
import warnings
import mock
import pytest
import numpy as np
import nrrd
import pandas as pd


from allensdk.core.mouse_connectivity_cache import MouseConnectivityCache
from allensdk.core.structure_tree import StructureTree
from allensdk.test_utilities.temp_dir import fn_temp_dir


@pytest.fixture(scope='function')
def mcc(fn_temp_dir):

    manifest_path = os.path.join(fn_temp_dir, 'manifest.json')
    return MouseConnectivityCache(manifest_file=manifest_path)


@pytest.fixture(scope='function')
def new_nodes():

    return [{'id': 0, 'structure_id_path': '/0/', 
             'color_hex_triplet': '000000', 'acronym': 'rt', 
             'name': 'root', 'structure_sets':[{'id': 1}, {'id': 4}]}]


@pytest.fixture(scope='function')
def old_nodes():

    return [{'id': 0, 'structure_id_path': '/0/', 
             'color_hex_triplet': '000000', 'acronym': 'rt', 
             'name': 'root', 'parent_structure_id': 12}]


@pytest.fixture(scope='function')
def experiments():

    return [{'num-voxels': 100, 'injection-volume': 99, 'sum': 98, 
             'name': 'foo', 'transgenic-line': 'most_creish', 
             'structure-id': 97,}]


@pytest.fixture(scope='function')
def unionizes():

    # note that I've mucked around with these values a bit
    return [{"hemisphere_id": 1, "id": 169991412, "is_injection": False,
             "max_voxel_density": 0.284863, "max_voxel_x": 7700, 
             "max_voxel_y": 6500, "max_voxel_z": 5000, 
             "normalized_projection_volume": 0.0, 
             "projection_density": 0.116754, "projection_energy": 30.7332,
             "projection_intensity": 263.231, "projection_volume": 0.0018718,
             "section_data_set_id": 166218353, "structure_id": 1,
             "sum_pixel_intensity": 99234900.0, "sum_pixels": 1308740.0,
             "sum_projection_pixel_intensity": 40221700.0,
             "sum_projection_pixels": 152800.0,
             "volume": 0.016032},
            {"hemisphere_id": 2, "id": 169991601, "is_injection": False,
             "max_voxel_density": 0.0614783, "max_voxel_x": 7500,
             "max_voxel_y": 4900, "max_voxel_z": 1700,
             "normalized_projection_volume": 0.0, 
             "projection_density": 0.0168009,
             "projection_energy": 1.96084, "projection_intensity": 116.71,
             "projection_volume": 0.00148144, 
             "section_data_set_id": 166218353, "structure_id": 60,
             "sum_pixel_intensity": 261941000.0, "sum_pixels": 7198050.0,
             "sum_projection_pixel_intensity": 14114200.0,
             "sum_projection_pixels": 120934.0, "volume": 0.0881761}]


def test_init(mcc, fn_temp_dir):

    manifest_path = os.path.join(fn_temp_dir, 'manifest.json')
    assert( os.path.exists(manifest_path) )


def test_get_annotation_volume(mcc, fn_temp_dir):

    eye = np.eye(100)
    path = os.path.join(fn_temp_dir, 'annotation', 'ccf_2016', 
                        'annotation_25.nrrd')

    mcc.api.retrieve_file_over_http = lambda a, b: nrrd.write(b, eye)
    obtained, _ = mcc.get_annotation_volume()

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_annotation_volume()

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) ) 
    assert( os.path.exists(path) )


def test_get_template_volume(mcc, fn_temp_dir):

    eye = np.eye(100)
    path = os.path.join(fn_temp_dir, 'average_template_25.nrrd')

    mcc.api.retrieve_file_over_http = lambda a, b: nrrd.write(b, eye)
    obtained, _ = mcc.get_template_volume()

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_template_volume()

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_projection_density(mcc, fn_temp_dir):

    eye = np.eye(100)
    eid = 123456789
    path = os.path.join(fn_temp_dir, 'experiment_{0}'.format(eid), 
                        'projection_density_25.nrrd')

    with mock.patch('allensdk.api.queries.grid_data_api.GridDataApi.'
                    'retrieve_file_over_http', 
                    new=lambda a, b, c: nrrd.write(c, eye)):
        obtained, _ = mcc.get_projection_density(eid)

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_projection_density(eid)

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_injection_density(mcc, fn_temp_dir):

    eye = np.eye(100)
    eid = 123456789
    path = os.path.join(fn_temp_dir, 'experiment_{0}'.format(eid), 
                        'injection_density_25.nrrd')

    with mock.patch('allensdk.api.queries.grid_data_api.GridDataApi.'
                    'retrieve_file_over_http', 
                    new=lambda a, b, c: nrrd.write(c, eye)):
        obtained, _ = mcc.get_injection_density(eid)

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_injection_density(eid)

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_injection_fraction(mcc, fn_temp_dir):

    eye = np.eye(100)
    eid = 123456789
    path = os.path.join(fn_temp_dir, 'experiment_{0}'.format(eid), 
                        'injection_fraction_25.nrrd')

    with mock.patch('allensdk.api.queries.grid_data_api.GridDataApi.'
                    'retrieve_file_over_http', 
                    new=lambda a, b, c: nrrd.write(c, eye)):
        obtained, _ = mcc.get_injection_fraction(eid)

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_injection_fraction(eid)

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_data_mask(mcc, fn_temp_dir):

    eye = np.eye(100)
    eid = 123456789
    path = os.path.join(fn_temp_dir, 'experiment_{0}'.format(eid), 
                        'data_mask_25.nrrd')

    with mock.patch('allensdk.api.queries.grid_data_api.GridDataApi.'
                    'retrieve_file_over_http', 
                    new=lambda a, b, c: nrrd.write(c, eye)):
        obtained, _ = mcc.get_data_mask(eid)

    mcc.api.retrieve_file_over_http = mock.MagicMock()
    mcc.get_data_mask(eid)

    mcc.api.retrieve_file_over_http.assert_not_called()
    assert( np.allclose(obtained, eye) )            
    assert( os.path.exists(path) )


def test_get_structure_tree(mcc, fn_temp_dir, new_nodes):

    path = os.path.join(fn_temp_dir, 'structures.json')

    with mock.patch('allensdk.api.queries.ontologies_api.'
                    'OntologiesApi.model_query', 
                    return_value=new_nodes) as p:

        obtained = mcc.get_structure_tree()

        mcc.get_structure_tree()
        p.assert_called_once()

    assert(obtained.node_ids()[0] == 0)
    
    cm_obt = obtained.get_colormap()
    assert(len(cm_obt[0]) == 3)

    assert( os.path.exists(path) )


def test_get_ontology(mcc, fn_temp_dir, old_nodes):

    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')

        with mock.patch('allensdk.api.queries.ontologies_api.'
                        'OntologiesApi.model_query', 
                        return_value=old_nodes) as p:

            mcc.get_ontology()
            mcc.get_ontology()

            p.assert_called_once()
            assert(len(c) == 6)


def test_get_structures(mcc, fn_temp_dir, old_nodes):

    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')

        with mock.patch('allensdk.api.queries.ontologies_api.'
                        'OntologiesApi.model_query', 
                        return_value=old_nodes) as p:

            obtained = mcc.get_structures()
            mcc.get_structures()

            p.assert_called_once()
            assert obtained['acronym'][0] == old_nodes[0]['acronym']
            assert len(c) == 2


def test_get_experiments(mcc, fn_temp_dir, experiments):

    file_path = os.path.join(fn_temp_dir, 'experiments.json')

    mcc.api.service_query = lambda a, parameters: experiments    
    obtained = mcc.get_experiments()

    mcc.api.service_query = mock.MagicMock()
    mcc.get_experiments()

    mcc.api.service_query.assert_not_called()
    assert os.path.exists(file_path)
    assert 'num_voxels' not in obtained[0]
    assert obtained[0]['transgenic-line'] == 'most_creish' 


def test_filter_experiments(mcc, fn_temp_dir, experiments):

    pass_line = mcc.filter_experiments(experiments, cre=True)
    fail_line = mcc.filter_experiments(experiments, cre=False)

    assert len(pass_line) == 1
    assert len(fail_line) == 0


def test_get_experiment_structure_unionizes(mcc, fn_temp_dir, unionizes):

    eid = 166218353
    path = os.path.join(fn_temp_dir, 'experiment_{0}'.format(eid), 
                        'structure_unionizes.csv')

    mcc.api.model_query = lambda *args, **kwargs: unionizes
    obtained = mcc.get_experiment_structure_unionizes(eid)

    mcc.api.model_query = mock.MagicMock()
    mcc.get_experiment_structure_unionizes(eid)

    mcc.api.model_query.assert_not_called()
    assert obtained.loc[0, 'projection_intensity'] == 263.231
    assert os.path.exists(path)


def test_filter_structure_unionizes(mcc, unionizes):

    obtained = mcc.filter_structure_unionizes(pd.DataFrame(unionizes), 
                                              hemisphere_ids=[1])

    assert obtained.loc[0, 'volume'] == 0.016032


def test_get_structure_unionizes(mcc, unionizes):

    mcc.get_experiment_structure_unionizes = \
        lambda *a, **k: pd.DataFrame(unionizes)
    obtained = mcc.get_structure_unionizes([1, 2, 3])

    assert obtained.shape[0] == 6


def test_get_projection_matrix(mcc):
    # yup

    unionizes = [{'experiment_id': 1, 
                  'structure_id': 2, 
                  'hemisphere_id': 1, 
                  'value': 30},
                 {'experiment_id': 1, 
                  'structure_id': 2, 
                  'hemisphere_id': 2, 
                  'value': 40},]

    mcc.get_structure_unionizes = lambda *a, **k: pd.DataFrame(unionizes)

    class FakeTree(object):
        def value_map(*a, **k):
            return {1: 'one', 2: 'two'}
    mcc.get_structure_tree = lambda *a, **k: FakeTree()

    obtained = mcc.get_projection_matrix([1], [2], [1, 2], ['value'])

    assert np.allclose(obtained['matrix'], np.array([[30, 40]]))
    assert np.array_equal([ii['label'] for ii in obtained['columns']], 
                          ['two-L', 'two-R'])


def test_get_reference_space(mcc, new_nodes):

    tree = StructureTree(StructureTree.clean_structures(new_nodes))
    mcc.get_structure_tree = lambda *a, **k: tree

    annot = np.arange(125).reshape((5, 5, 5))
    mcc.get_annotation_volume = lambda *a, **k: (annot, 'foo')

    rsp_obt = mcc.get_reference_space()

    assert( np.allclose(rsp_obt.resolution, [25, 25, 25]) )
    assert( np.allclose( rsp_obt.annotation, annot ) ) 


def test_get_structure_mask(mcc, fn_temp_dir):

    class FakeTree(object):
        def descendant_ids(self, list_of_things):
            return [list_of_things]
    mcc.get_structure_tree = lambda *a, **k: FakeTree()

    annot = np.arange(125).reshape((5, 5, 5))
    mcc.get_annotation_volume = lambda *a, **k: (annot, 'foo')

    path = os.path.join(fn_temp_dir, 'annotation', 'ccf_2016', 'structure_masks', 
                        'resolution_25', 'structure_{0}.nrrd'.format(12))

    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')

        mask, _ = mcc.get_structure_mask(12)

        # also make sure we can do this for pd.Series input for backwards compatibility
        mask, _ = mcc.get_structure_mask(pd.Series([12]))

    assert( mask.sum() == 1 )
    #assert( len(c) == 2 )
    assert( os.path.exists(path) )

    with pytest.raises(ValueError):
        mask, _ = mcc.get_structure_mask("fish")


def test_make_structure_mask(mcc):

    annot = np.arange(125).reshape((5, 5, 5))
    sids = [0, 1, 2, 3, 4]

    with warnings.catch_warnings(record=True) as c:
        warnings.simplefilter('always')

        mask = mcc.make_structure_mask(sids, annot)

    #assert(len(c) == 1)
    assert mask.sum() == 5

@pytest.mark.parametrize('inp,fails', [(1, False), 
                                        (pd.Series([2]), False), 
                                        ('qwerty', True)])
def test_validate_structure_id(inp, fails):

    if fails:
        with pytest.raises(ValueError) as exc:
            MouseConnectivityCache.validate_structure_id(inp)
    else:
        out = MouseConnectivityCache.validate_structure_id(inp)
        assert( out == int(inp) )


@pytest.mark.parametrize('inp,fails', [([1, 2, 3], False), 
                                        ([pd.Series([2]), pd.Series([3])], False), 
                                        (['qwerty', 1], True)])
def test_validate_structure_ids(inp, fails):

    if fails:
        with pytest.raises(ValueError) as exc:
            MouseConnectivityCache.validate_structure_ids(inp)
    else:
        out = MouseConnectivityCache.validate_structure_ids(inp)
        assert( out == map(int, inp) )



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np
from pkg_resources import resource_filename  # @UnresolvedImport
from allensdk.core.brain_observatory_nwb_data_set import BrainObservatoryNwbDataSet, si
import pytest
import os


NWB_FLAVORS = []


if 'TEST_NWB_FILES' in os.environ:
    nwb_list_file = os.environ['TEST_NWB_FILES']
else:
    nwb_list_file = resource_filename(__name__, 'nwb_files.txt')

if nwb_list_file == 'skip':
    NWB_FLAVORS = []
else:
    with open(nwb_list_file, 'r') as f:
        NWB_FLAVORS = [l.strip() for l in f]


@pytest.fixture(params=NWB_FLAVORS)
def data_set(request):
    data_set = BrainObservatoryNwbDataSet(request.param)

    return data_set


def test_acceptance(data_set):
    data_set.get_cell_specimen_ids()
    data_set.get_session_type()
    data_set.get_metadata()
    data_set.get_running_speed()
    data_set.get_motion_correction()


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_roi_ids(data_set):
    ids = data_set.get_roi_ids()
    assert len(ids) == len(data_set.get_cell_specimen_ids())

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_metadata(data_set):
    md = data_set.get_metadata()

    valid_fields = [ 'genotype', 'cre_line', 'imaging_depth_um', 'ophys_experiment_id', 'experiment_container_id',
                     'session_start_time', 'age_days', 'device', 'device_name', 'pipeline_version', 'sex',
                     'targeted_structure', 'excitation_lambda', 'indicator', 'fov', 'session_type', 'specimen_name' ]

    invalid_fields = [ 'imaging_depth', 'age', 'device_string', 'generated_by' ]

    for field in valid_fields:
        assert md[field] is not None

    for field in invalid_fields:
        assert field not in md
    


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_cell_specimen_indices(data_set):
    inds = data_set.get_cell_specimen_indices([])
    assert len(inds) == 0

    ids = data_set.get_cell_specimen_ids()

    inds = data_set.get_cell_specimen_indices(ids)
    assert np.all(np.array(inds) == np.arange(len(inds)))

    inds = data_set.get_cell_specimen_indices([ids[0]])
    assert inds[0] == 0


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_fluorescence_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_fluorescence_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_fluorescence_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_fluorescence_traces([ids[0]])
    assert traces.shape[0] == 1


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_neuropil_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_neuropil_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_neuropil_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_neuropil_traces([ids[0]])
    assert traces.shape[0] == 1


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_dff_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_dff_traces()
    # assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_dff_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_dff_traces([ids[0]])
    assert traces.shape[0] == 1

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_neuropil_r(data_set):

    ids = data_set.get_cell_specimen_ids()
    r = data_set.get_neuropil_r()
    assert len(ids) == len(r)

    r = data_set.get_neuropil_r(ids)
    assert len(ids) == len(r)

    short_list = [ids[0]]
    r = data_set.get_neuropil_r(short_list)
    assert len(short_list) == len(r)

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_corrected_fluorescence_traces(data_set):
    ids = data_set.get_cell_specimen_ids()

    timestamps, traces = data_set.get_corrected_fluorescence_traces()
    assert len(timestamps) == traces.shape[1]
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_corrected_fluorescence_traces(ids)
    assert len(ids) == traces.shape[0]

    timestamps, traces = data_set.get_corrected_fluorescence_traces([ids[0]])
    assert traces.shape[0] == 1


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_roi_mask(data_set):
    ids = data_set.get_cell_specimen_ids()
    roi_masks = data_set.get_roi_mask()
    assert len(ids) == len(roi_masks)

    max_projection = data_set.get_max_projection()
    for roi_mask in roi_masks:
        mask = roi_mask.get_mask_plane()
        assert mask.shape[0] == max_projection.shape[0]
        assert mask.shape[1] == max_projection.shape[1]

    roi_masks = data_set.get_roi_mask([ids[0]])
    assert len(roi_masks) == 1


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_roi_mask_array(data_set):
    ids = data_set.get_cell_specimen_ids()
    arr = data_set.get_roi_mask_array()
    assert arr.shape[0] == len(ids)

    arr = data_set.get_roi_mask_array([ids[0]])
    assert arr.shape[0] == 1

    try:
        arr = data_set.get_roi_mask_array([0])
    except ValueError as e:
        assert str(e).startswith("Cell specimen not found")


@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_stimulus_epoch_table(data_set):

    summary_df = data_set.get_stimulus_epoch_table()

    session_type = data_set.get_session_type()
    if session_type == si.THREE_SESSION_A or si.THREE_SESSION_C:
        assert len(summary_df) == 7
    elif session_type == si.THREE_SESSION_B:
        assert len(summary_df) == 8
    elif session_type == si.THREE_SESSION_C2:
        assert len(summary_df) == 10
    else:
        raise NotImplementedError('Code not tested for session of type: %s' % session_type)

@pytest.mark.skipif(not os.path.exists('/projects/neuralcoding'),
                    reason="test NWB file not available")
def test_get_stimulus_table_master(data_set):

    master_df = data_set.get_stimulus_table('master')

    session_type = data_set.get_session_type()
    if session_type == si.THREE_SESSION_A:
        assert len(master_df) == 45629
    elif session_type == si.THREE_SESSION_B:
        assert len(master_df) == 20951
    elif session_type == si.THREE_SESSION_C:
        assert len(master_df) == 26882
    elif session_type == si.THREE_SESSION_C2:
        assert len(master_df) == 29398
    else:
        raise NotImplementedError('Code not tested for session of type: %s' % session_type)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from ..api import Api
import warnings


class RmaApi(Api):
    '''
    See: `RESTful Model Access (RMA) <http://help.brain-map.org/display/api/RESTful+Model+Access+%28RMA%29>`_
    '''
    MODEL = 'model::'
    PIPE = 'pipe::'
    SERVICE = 'service::'
    CRITERIA = 'rma::criteria'
    INCLUDE = 'rma::include'
    OPTIONS = 'rma::options'
    ORDER = 'order'
    NUM_ROWS = 'num_rows'
    ALL = 'all'
    START_ROW = 'start_row'
    COUNT = 'count'
    ONLY = 'only'
    EXCEPT = 'except'
    EXCPT = 'excpt'
    TABULAR = 'tabular'
    DEBUG = 'debug'
    PREVIEW = 'preview'
    TRUE = 'true'
    FALSE = 'false'
    IS = '$is'
    EQ = '$eq'

    def __init__(self, base_uri=None):
        super(RmaApi, self).__init__(base_uri)

    def build_query_url(self,
                        stage_clauses,
                        fmt='json'):
        '''Combine one or more RMA query stages into a single RMA query.

        Parameters
        ----------
        stage_clauses : list of strings
            subqueries
        fmt : string, optional
            json (default), xml, or csv

        Returns
        -------
        string
            complete RMA url
        '''
        if not type(stage_clauses) is list:
            stage_clauses = [stage_clauses]

        url = ''.join([
            self.rma_endpoint,
            '/query.',
            fmt,
            '?q=',
            ','.join(stage_clauses)])

        return url

    def model_stage(self,
                    model,
                    **kwargs):
        '''Construct a model stage of an RMA query string.

        Parameters
        ----------
        model : string
            The top level data type
        filters : dict
            key, value comparisons applied to the top-level model to narrow the results.
        criteria : string
            raw RMA criteria clause to choose what object are returned
        include : string
            raw RMA include clause to return associated objects
        only : list of strings, optional
            to be joined into an rma::options only filter to limit what data is returned
        except : list of strings, optional
            to be joined into an rma::options except filter to limit what data is returned
        tabular : list of string, optional
            return columns as a tabular data structure rather than a nested tree.
        count : boolean, optional
            False to skip the extra database count query.
        debug : string, optional
            'true', 'false' or 'preview'
        num_rows : int or string, optional
            how many database rows are returned (may not correspond directly to JSON tree structure)
        start_row : int or string, optional
            which database row is start of returned data  (may not correspond directly to JSON tree structure)


        Notes
        -----
        See `RMA Path Syntax <http://help.brain-map.org/display/api/RMA+Path+Syntax#RMAPathSyntax-DoubleColonforAxis>`_
        for a brief overview of the normalized RMA syntax.
        Normalized RMA syntax differs from the legacy syntax
        used in much of the RMA documentation.
        Using the &debug=true option with an RMA URL will include debugging information in the
        response, including the normalized query.
        '''
        clauses = [RmaApi.MODEL + model]

        filters = kwargs.get('filters', None)

        if filters is not None:
            clauses.append(self.filters(filters))

        criteria = kwargs.get('criteria', None)

        if criteria is not None:
            clauses.append(',')
            clauses.append(RmaApi.CRITERIA)
            clauses.append(',')
            clauses.extend(criteria)

        include = kwargs.get('include', None)

        if include is not None:
            clauses.append(',')
            clauses.append(RmaApi.INCLUDE)
            clauses.append(',')
            clauses.extend(include)

        options_clause = self.options_clause(**kwargs)

        if options_clause != '':
            clauses.append(',')
            clauses.append(options_clause)

        stage = ''.join(clauses)

        return stage

    def pipe_stage(self,
                   pipe_name,
                   parameters):
        '''Connect model and service stages via their JSON responses.

        Notes
        -----
        See: `Service Pipelines <http://help.brain-map.org/display/api/Service+Pipelines>`_
        and
        `Connected Services and Pipes <http://help.brain-map.org/display/api/Connected+Services+and+Pipes>`_
        '''
        clauses = [RmaApi.PIPE + pipe_name]

        clauses.append(self.tuple_filters(parameters))

        stage = ''.join(clauses)

        return stage

    def service_stage(self,
                      service_name,
                      parameters=None):
        '''Construct an RMA query fragment to send a request to a connected service.

        Parameters
        ----------
        service_name : string
            Name of a documented connected service.
        parameters : dict
            key-value pairs as in the online documentation.

        Notes
        -----
        See: `Service Pipelines <http://help.brain-map.org/display/api/Service+Pipelines>`_
        and
        `Connected Services and Pipes <http://help.brain-map.org/display/api/Connected+Services+and+Pipes>`_
        '''
        clauses = [RmaApi.SERVICE + service_name]

        if parameters is not None:
            clauses.append(self.tuple_filters(parameters))

        stage = ''.join(clauses)

        return stage

    def model_query(self, *args, **kwargs):
        '''Construct and execute a model stage of an RMA query string.

        Parameters
        ----------
        model : string
            The top level data type
        filters : dict
            key, value comparisons applied to the top-level model to narrow the results.
        criteria : string
            raw RMA criteria clause to choose what object are returned
        include : string
            raw RMA include clause to return associated objects
        only : list of strings, optional
            to be joined into an rma::options only filter to limit what data is returned
        except : list of strings, optional
            to be joined into an rma::options except filter to limit what data is returned
        excpt : list of strings, optional
            synonym for except parameter to avoid a reserved word conflict.
        tabular : list of string, optional
            return columns as a tabular data structure rather than a nested tree.
        count : boolean, optional
            False to skip the extra database count query.
        debug : string, optional
            'true', 'false' or 'preview'
        num_rows : int or string, optional
            how many database rows are returned (may not correspond directly to JSON tree structure)
        start_row : int or string, optional
            which database row is start of returned data  (may not correspond directly to JSON tree structure)


        Notes
        -----
        See `RMA Path Syntax <http://help.brain-map.org/display/api/RMA+Path+Syntax#RMAPathSyntax-DoubleColonforAxis>`_
        for a brief overview of the normalized RMA syntax.
        Normalized RMA syntax differs from the legacy syntax
        used in much of the RMA documentation.
        Using the &debug=true option with an RMA URL will include debugging information in the
        response, including the normalized query.
        '''
        return self.json_msg_query(
            self.build_query_url(
                self.model_stage(*args, **kwargs)))

    def service_query(self, *args, **kwargs):
        '''Construct and Execute a single-stage RMA query
        to send a request to a connected service.

        Parameters
        ----------
        service_name : string
            Name of a documented connected service.
        parameters : dict
            key-value pairs as in the online documentation.

        Notes
        -----
        See: `Service Pipelines <http://help.brain-map.org/display/api/Service+Pipelines>`_
        and
        `Connected Services and Pipes <http://help.brain-map.org/display/api/Connected+Services+and+Pipes>`_
        '''
        return self.json_msg_query(
            self.build_query_url(
                self.service_stage(*args, **kwargs)))

    def options_clause(self, **kwargs):
        '''build rma:: options clause.

        Parameters
        ----------
        only : list of strings, optional
        except : list of strings, optional
        tabular : list of string, optional
        count : boolean, optional
        debug : string, optional
            'true', 'false' or 'preview'
        num_rows : int or string, optional
        start_row : int or string, optional
        '''
        clause = ''
        options_params = []

        only = kwargs.get(RmaApi.ONLY, None)

        if only is not None:
            options_params.append(
                self.only_except_tabular_clause(RmaApi.ONLY,
                                                only))

        # handle alternate 'except' spelling to avoid reserved word conflict
        excpt = kwargs.get(RmaApi.EXCEPT, None)
        excpt2 = kwargs.get(RmaApi.EXCPT, None)
        
        if excpt is not None and excpt2 is not None:
            warnings.warn('excpt and except options should not be used together',
                          Warning)
        elif excpt2 is not None:
            excpt = excpt2 

        if excpt is not None:
            options_params.append(
                self.only_except_tabular_clause(RmaApi.EXCEPT,
                                                excpt))

        tabular = kwargs.get(RmaApi.TABULAR, None)

        if tabular is not None:
            options_params.append(
                self.only_except_tabular_clause(RmaApi.TABULAR,
                                                tabular))

        num_rows = kwargs.get(RmaApi.NUM_ROWS, None)

        if num_rows is not None:
            if num_rows == RmaApi.ALL:
                options_params.append("[%s$eq'all']" % (RmaApi.NUM_ROWS))
            else:
                options_params.append('[%s$eq%d]' % (RmaApi.NUM_ROWS,
                                                     num_rows))

        start_row = kwargs.get(RmaApi.START_ROW, None)

        if start_row is not None:
            options_params.append('[%s$eq%d]' % (RmaApi.START_ROW,
                                                 start_row))

        order = kwargs.get(RmaApi.ORDER, None)

        if order is not None:
            options_params.append(self.order_clause(order))

        debug = kwargs.get(RmaApi.DEBUG, None)

        if debug is not None:
            options_params.append(self.debug_clause(debug))

        cnt = kwargs.get(RmaApi.COUNT, None)

        if cnt is not None:
            if cnt is True or cnt == 'true':
                options_params.append('[%s$eq%s]' % (RmaApi.COUNT,
                                                     RmaApi.TRUE))
            elif cnt is False or cnt == 'false':
                options_params.append('[%s$eq%s]' % (RmaApi.COUNT,
                                                     RmaApi.FALSE))
            else:
                pass

        if len(options_params) > 0:
            clause = RmaApi.OPTIONS + ''.join(options_params)

        return clause

    def only_except_tabular_clause(self, filter_type, attribute_list):
        '''Construct a clause to filter which attributes are returned
        for use in an rma::options clause.

        Parameters
        ----------
        filter_type : string
            'only', 'except', or 'tabular'
        attribute_list : list of strings
            for example ['acronym', 'products.name', 'structure.id']

        Returns
        -------
        clause : string
            The query clause for inclusion in an RMA query URL.

        Notes
        -----
        The title of tabular columns can be set by adding '+as+<title>'
        to the attribute.
        The tabular filter type requests a response that is row-oriented
        rather than a nested structure.
        Because of this, the tabular option can mask the lazy query behavior
        of an rma::include clause.
        The tabular option does not mask the inner-join behavior of an rma::include
        clause.
        The tabular filter is required for .csv format RMA requests.
        '''
        clause = ''

        if attribute_list is not None:
            clause = '[%s$eq%s]' % (filter_type,
                                    ','.join(attribute_list))

        return clause

    def order_clause(self, order_list=None):
        '''Construct a debug clause for use in an rma::options clause.

        Parameters
        ----------
        order_list : list of strings
            for example ['acronym', 'products.name+asc', 'structure.id+desc']

        Returns
        -------
        clause : string
            The query clause for inclusion in an RMA query URL.

        Notes
        -----
        Optionally adding '+asc' (default) or '+desc' after an attribute
        will change the sort order.
        '''
        clause = ''

        if order_list is not None:
            clause = '[order$eq%s]' % (','.join(order_list))

        return clause

    def debug_clause(self, debug_value=None):
        '''Construct a debug clause for use in an rma::options clause.
        Parameters
        ----------
        debug_value : string or boolean
            True, False, None (default) or 'preview'

        Returns
        -------
        clause : string
            The query clause for inclusion in an RMA query URL.

        Notes
        -----
        True will request debugging information in the response.
        False will request no debugging information.
        None will return an empty clause.
        'preview' will request debugging information without the query being run.

        '''
        clause = ''

        if debug_value is None:
            clause = ''
        if debug_value is True or debug_value == 'true':
            clause = '[debug$eqtrue]'
        elif debug_value is False or debug_value == 'false':
            clause = '[debug$eqfalse]'
        elif debug_value == 'preview':
            clause = "[debug$eq'preview']"

        return clause

    # TODO: deprecate for something that can preserve order
    def filters(self, filters):
        '''serialize RMA query filter clauses.

        Parameters
        ----------
        filters : dict
            keys and values for narrowing a query.

        Returns
        -------
        string
            filter clause for an RMA query string.
        '''
        filters_builder = []

        for (key, value) in filters.items():
            filters_builder.append(self.filter(key, value))

        return ''.join(filters_builder)

    # TODO: this needs to be more rigorous.
    def tuple_filters(self, filters):
        '''Construct an RMA filter clause.

        Notes
        -----

        See `RMA Path Syntax - Square Brackets for Filters <http://help.brain-map.org/display/api/RMA+Path+Syntax#RMAPathSyntax-SquareBracketsforFilters>`_ for additional documentation.
        '''
        filters_builder = []

        for filt in sorted(filters):
            if filt[-1] is None:
                continue
            if len(filt) == 2:
                val = filt[1]
                if type(val) is list:
                    val_array = []
                    for v in val:
                        if type(v) is str:
                            val_array.append(v)
                        else:
                            val_array.append(str(v))
                    val = ','.join(val_array)
                    filters_builder.append("[%s$eq%s]" % (filt[0], val))
                elif type(val) is int:
                    filters_builder.append("[%s$eq%d]" % (filt[0], val))
                elif type(val) is bool:
                    if val:
                        filters_builder.append("[%s$eqtrue]" % (filt[0]))
                    else:
                        filters_builder.append("[%s$eqfalse]" % (filt[0]))
                elif type(val) is str:
                    filters_builder.append("[%s$eq%s]" % (filt[0], filt[1]))
            elif len(filt) == 3:
                filters_builder.append("[%s%s%s]" % (filt[0],
                                                     filt[1],
                                                     str(filt[2])))

        return ''.join(filters_builder)

    def quote_string(self, the_string):
        '''Wrap a clause in single quotes.

        Parameters
        ----------
        the_string : string
            a clause to be included in an rma query that needs to be quoted

        Returns
        -------
        string
            input wrapped in single quotes
        '''
        return ''.join(["'", the_string, "'"])

    def filter(self, key, value):
        '''serialize a single RMA query filter clause.

        Parameters
        ----------
        key : string
            keys for narrowing a query.
        value : string
            value for narrowing a query.

        Returns
        -------
        string
            a single filter clause for an RMA query string.
        '''
        return "".join(['[',
                        key,
                        RmaApi.EQ,
                        str(value),
                        ']'])

    def build_schema_query(self, clazz=None, fmt='json'):
        '''Build the URL that will fetch the data schema.

        Parameters
        ----------
        clazz : string, optional
            Name of a specific class or None (default).
        fmt : string, optional
            json (default) or xml

        Returns
        -------
        url : string
            The constructed URL

        Notes
        -----
        If a class is specified, only the schema information for that class
        will be requested, otherwise the url requests the entire schema.
        '''
        if clazz is not None:
            class_clause = '/' + clazz
        else:
            class_clause = ''

        url = ''.join([self.rma_endpoint,
                       class_clause,
                       '.',
                       fmt])

        return url

    def get_schema(self, clazz=None):
        '''Retrieve schema information.'''
        schema_data = self.do_query(self.build_schema_query,
                                    self.read_data,
                                    clazz)

        return schema_data



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import numpy as np

import pandas as pd
import math
try:
    xrange
except:
    from past.builtins import xrange
from matplotlib.colors import LinearSegmentedColormap
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.collections import PatchCollection, LineCollection
import matplotlib.transforms as mxfms
import matplotlib.colors as mcolors
import skimage.transform

DEFAULT_COLOR_MAP = LinearSegmentedColormap.from_list('default', [[.7,0,.7,0.0],[.7,0,0,1]])
DEFAULT_MEAN_RESP_COLOR_MAP = LinearSegmentedColormap.from_list('default', [[0.0,0.0,0.5,0.0],[0.0,0.0,0.5,1]])
DEFAULT_AXIS_COLOR = (0.8, 0.8, 0.8)
DEFAULT_LABEL_COLOR = (0.8, 0.8, 0.8)
LSN_ON_COLOR_MAP = LinearSegmentedColormap.from_list('default', [[.7,0,.7,0.0],[.7,0,0,1]])
LSN_OFF_COLOR_MAP = LinearSegmentedColormap.from_list('default', [[0.0,0.7,.7,0.0],[0,0,0.7,1]])
HEX_POSITIONS = []


def polar_to_xy(angles, radius):
    """ Convert an array of angles (in radians) and a radius in polar coordinates 
    to an array of x,y coordinates.  
    """

    x = radius*np.cos(angles)
    y = radius*np.sin(angles)
    return np.array([x,y]).T


def polar_linspace(radius, start_angle, stop_angle, num, endpoint=False, degrees=True):
    """ Evenly distributed list of x,y coordinates from an input range of angles 
    and a radius in polar coordinates.  
    """
    angles = np.linspace(start_angle, stop_angle, num=num, endpoint=endpoint)

    if degrees is True:
        angles *= np.pi / 180.0

    return polar_to_xy(angles, radius)


def spiral_trials(radii, x=0.0, y=0.0):
    radii = np.array(radii)
    circles = []

    if radii.size > 0:
        spiral = hex_pack(radii[0], len(radii))

        for i,radius in enumerate(radii):
            circles.append(mpatches.Circle((spiral[i][0], spiral[i][1]), radii[i]))
        
    pos_xfm = mxfms.Affine2D().translate(x,y)

    collection = PatchCollection(circles)
    collection.set_transform(pos_xfm)

    return collection


def spiral_trials_polar(r, theta, radii, offset=None):
    if offset is None:
        offset = [0,0]

    collection = spiral_trials(radii, r + offset[0], offset[1])

    rot_xfm = mxfms.Affine2D().rotate(theta)
    collection.set_transform(collection.get_transform() + rot_xfm)

    return collection


def angle_lines(angles, inner_radius, outer_radius):
    inner_pos = polar_to_xy(angles, inner_radius)
    outer_pos = polar_to_xy(angles, outer_radius)

    segments = np.array(list(zip(inner_pos, outer_pos)))

    return LineCollection(segments)


def radial_arcs(rs, start_theta, end_theta):
    arcs = []

    for r in rs:
        arcs.append(mpatches.Arc((0,0), 2*r, 2*r,
                                 theta1=start_theta*180.0/np.pi,
                                 theta2=end_theta*180.0/np.pi))
                                  

    return PatchCollection(arcs)

def rings_in_hex_pack(ct):
    return np.ceil((-3.0 + np.sqrt(9.0 - 12.0*(1.0 - ct))) / 6.0 + 1.0)

def radial_circles(rs):
    circles = [ mpatches.Circle((0,0), r) for r in rs ]

    return PatchCollection(circles)

def reset_hex_pack():
    global HEX_POSITIONS
    HEX_POSITIONS = []
    
def hex_pack(radius, n):
    global HEX_POSITIONS

    if len(HEX_POSITIONS) < n:
        HEX_POSITIONS = build_hex_pack(n)

    return HEX_POSITIONS[:n]*radius*2.0

def build_hex_pack(n):
    pos = []
    sq32 = math.sqrt(3.0) / 2.0
    
    N = 1
    
    vs = [ [-0.5, -sq32], [-1.0, 0.0], [-0.5, sq32], [0.5, sq32], [1, 0], [0.5, -sq32] ]    
    pos.append([0,0])
    while len(pos) < n:
        layer_pos = [  ] 
        
        for i,v in enumerate(vs):
            x = - N * v[1] * sq32
            y = N * v[0] * sq32
        
            if N % 2 == 1:
                x -= 0.5 * v[0]
                y -= 0.5 * v[1]
                
            layer_pos.append([])
            layer_pos[i].append([x,y])
            mag = 1
            sign = 1
                      
            for j in xrange(N-1):
                x += v[0] * mag * sign
                y += v[1] * mag * sign
                mag += 1
                sign = -sign
                layer_pos[i].append([x,y])
        
        for j in range(N):
            for i in range(len(vs)):
                if j < len(layer_pos[i]):
                    pos.append(layer_pos[i][j])
        N+=1        

    return np.array(pos)


def polar_line_circles(radii, theta, start_r=0):
    circles = [ mpatches.Circle( (0,0), radii[0] ) ]
    
    line_xfm = mxfms.Affine2D().translate(start_r,0).rotate(theta)
    
    x = 0
    for ri in range(1, len(radii)):
        x += radii[ri-1] + radii[ri]
        
        circles.append(mpatches.Circle( (x,0), radii[ri] ))
        
    collection = PatchCollection(circles)
    collection.set_transform(line_xfm)
    
    return collection


def wedge_ring(N, inner_radius, outer_radius, start=0, stop=360):
    degs = np.linspace(start, stop, N+1, endpoint=True)
    wedges = []

    if stop > start:
        for i in range(len(degs)-1):
            wedges.append( mpatches.Wedge( (0,0), outer_radius, degs[i], degs[i+1], width=outer_radius-inner_radius ) )
    else:
        for i in range(1,len(degs)):
            wedges.append( mpatches.Wedge( (0,0), outer_radius, degs[i], degs[i-1], width=outer_radius-inner_radius ) )

    return PatchCollection(wedges)


def add_angle_labels(ax, angles, labels, radius, color=None, fontdict=None, offset=0.05):
    angle_pos = polar_to_xy(angles, radius)

    for i in range(len(angle_pos)):
        xy = angle_pos[i,:]
        u = xy + xy / np.linalg.norm(xy) * offset
        ax.text(u[0], u[1],
                labels[i], color=color, 
                horizontalalignment='center', 
                verticalalignment='center',
                fontdict=fontdict)    
    

def add_arrow(ax, radius, start_angle, end_angle, color=None, width=18.0):
    if color is None:
        color = DEFAULT_LABEL_COLOR

    fig = ax.get_figure()
    size = fig.get_size_inches()
    dpi = fig.get_dpi()
    mutation_scale = size[0] * dpi / 500.0 * width

    d_angle = end_angle - start_angle

    start_pos = (radius * np.cos(start_angle), radius * np.sin(start_angle))
    end_pos = (radius * np.cos(end_angle), radius * np.sin(end_angle))
             
    connstyle = mpatches.ConnectionStyle.Angle3(angleA=0, angleB=(d_angle*180.0/np.pi))
    arrowstyle = mpatches.ArrowStyle.Simple(tail_width=0.33, head_length=0.66, head_width=1.0)
    ax.add_patch(mpatches.FancyArrowPatch(posA=start_pos, posB=end_pos,
                                          arrowstyle=arrowstyle,
                                          connectionstyle=connstyle,
                                          facecolor=color,
                                          linewidth=0,
                                          mutation_scale=mutation_scale))

def make_pincushion_plot(data, trials, on, nrows, ncols, clim=None, color_map=None, radius=None):
    if radius is None:
        max_sweeps = 0
        for sweeps in trials.itervalues():
            max_sweeps = max(max_sweeps, len(sweeps[0]))

        rings = rings_in_hex_pack(max_sweeps)
        radius = 0.5 / (2.0 * rings - 1.0)

    if clim is None:
        clim = [ data.min(), data.max() ]

    if color_map is None:
        color_map = LSN_ON_COLOR_MAP if on else LSN_OFF_COLOR_MAP

    ax = plt.gca()
    for (col,row,on_state), sweeps in trials.iteritems():
        if on_state != on:
            continue

        valid_sweeps = sweeps[0][sweeps[0] < data.size]
        responses = np.sort(data[valid_sweeps])[::-1]
        responses = responses[responses >= clim[0]]
        
        if responses.size > 0:
            coll = spiral_trials(np.ones(responses.shape)*radius, col+0.5, row+0.5)
            coll.set_transform(coll.get_transform() + ax.transData)
            coll.set_array(responses)
            coll.set_cmap(color_map)
            coll.set_clim(clim)
            coll.set_linewidths(0)
            ax.add_collection(coll)

    ax.set_ylim((0,nrows))
    ax.set_xlim((0,ncols))

    

class PolarPlotter( object ):
    DIR_CW = -1
    DIR_CCW = 1

    def __init__(self, 
                 direction=DIR_CW,
                 angle_start=0,
                 circle_scale=1.1,
                 inner_radius=None,
                 plot_center=(0.0,0.0),
                 plot_scale=0.9):

        self.plot_scale = plot_scale
        self.plot_center = plot_center

        self.angle_transform = np.vectorize(lambda x: ((x + angle_start)*direction)*np.pi/180.0)
        self.inner_radius = inner_radius
        self.circle_scale = circle_scale

    def finalize(self):
        ax = plt.gca()
        fig = plt.gcf()
        figsize = fig.get_size_inches()

        aspect = figsize[0] / figsize[1]
        w = 2.0 / self.plot_scale 
        h = w / aspect

        bounds = ( self.plot_center[0] - w*.5,
                   self.plot_center[0] + w*.5,
                   self.plot_center[1] - h*.5,
                   self.plot_center[1] + h*.5 )

        ax.set_xlim(bounds[0], bounds[1])
        ax.set_ylim(bounds[2], bounds[3])

        plt.subplots_adjust(left=0,right=1,bottom=0,top=1)

    @classmethod
    def _clim(self, clim, data):

        if clim is None:
            clim = [ data.min(), data.max() ]

            if clim[0] == clim[1]:
                clim[0] = 0
            if clim[0] == clim[1]:
                clim[1] = 1

        return clim


class TrackPlotter( PolarPlotter ):
    def __init__(self, 
                 direction=PolarPlotter.DIR_CW,
                 angle_start=270.0,
                 inner_radius=.45,
                 ring_length=None, 
                 *args, **kwargs):
        super(TrackPlotter, self).__init__(direction=direction,
                                           angle_start=angle_start,
                                           inner_radius=inner_radius, 
                                           *args, **kwargs)

        self.ring_length = ring_length

    def show_arrow(self, color=None):
        start, end = self.angle_transform([0.0, 40.0])
        add_arrow(plt.gca(), self.inner_radius * .85, start, end, color)

    def plot(self, data,
             clim=None,
             cmap=DEFAULT_COLOR_MAP,
             mean_cmap=DEFAULT_MEAN_RESP_COLOR_MAP,
             norm=None):

        ax = plt.gca()

        clim = self._clim(clim, data)
        if self.ring_length:
            data = skimage.transform.resize(data.astype(np.float64), (data.shape[0], self.ring_length))
   
        data_mean = data.mean(axis=0)
        data = np.vstack((data, data_mean))
        
        radii = np.linspace(self.inner_radius, 1.0, data.shape[0]+2)
        start,stop = self.angle_transform([0,360])*180.0/np.pi

        if norm is None:
            norm = mcolors.PowerNorm(0.5, vmin=clim[0], vmax=clim[1], clip=True)

        for i, row_data in enumerate(data):
            inner_radius = radii[i]

            if i < data.shape[0] - 1:
                outer_radius = radii[i+1]
                ring_cmap = cmap
            else:
                outer_radius = radii[i+2]
                ring_cmap = mean_cmap


            wedges = wedge_ring(len(row_data), 
                                inner_radius, outer_radius,
                                start=start, stop=stop)

            wedges.set_array(row_data)
            #wedges.set_clim(clim)
            wedges.set_cmap(ring_cmap)
            wedges.set_norm(norm)
            wedges.set_edgecolors((0,0,0,0))

            ax.add_collection(wedges)

        self.finalize()
        

class CoronaPlotter( PolarPlotter ):
    def __init__(self, 
                 angle_start=270, 
                 plot_scale=1.2, 
                 inner_radius=.3,
                 *args, **kwargs):
        super(CoronaPlotter, self).__init__(inner_radius=inner_radius, angle_start=angle_start, plot_scale=plot_scale, *args, **kwargs)

        self.categories = None
        self.cat_idx_map = None

    def infer_dims(self, category_data):
        self.set_dims(np.sort(np.unique(category_data)))

    def set_dims(self, categories):
        self.categories = categories
        self.cat_idx_map = dict(zip(categories, range(len(categories))))

    def show_arrow(self, color=None):
        start, end = self.angle_transform([0.0, 40.0])
        add_arrow(plt.gca(), self.inner_radius * .85, start, end, color)

    def plot(self, category_data, 
             data=None,
             clim=None, 
             cmap=DEFAULT_COLOR_MAP):
             
        ax = plt.gca()

        if self.categories is None:
            self.infer_dims(category_data)

        if data is None:
            data = np.ones(len(category_data))
    
        clim = self._clim(clim, data)

        num_cats = len(self.categories)
        hth = 180.0 / num_cats
        degs = np.linspace(hth, 360.0-hth, num_cats)
        degs = self.angle_transform(degs)
        circle_radius = self.inner_radius * abs(np.sin((degs[1] - degs[0]) * .5))
        
        radii = np.ones(len(data)) * circle_radius * self.circle_scale

        df = pd.DataFrame({ 'category': category_data })
        gb = df.groupby(['category'])

        for category, trials in gb.groups.iteritems():
            idx = self.cat_idx_map[category]
            order = np.argsort(data[trials])[::-1]
            trial_order = np.array(trials)[order]

            circles = polar_line_circles(radii[trial_order],
                                         degs[idx], 
                                         self.inner_radius)                                         

            circles.set_transform(circles.get_transform() + ax.transData)
            circles.set_array(data[trial_order])
            circles.set_cmap(cmap)
            circles.set_clim(clim)
            circles.set_edgecolors((0,0,0,0))

            ax.add_collection(circles)

        self.finalize()

class FanPlotter( PolarPlotter ):
    def __init__(self, group_scale=0.9, *args, **kwargs):
        super(FanPlotter, self).__init__(*args, **kwargs)

        self.group_scale = group_scale

        self.angles = None
        self.xangles = None
        self.angle_map = None

        self.rs = None
        self.radii = None
        self.r_radius_map = None

        self.groups = None
        self.group_offsets = None
        self.group_offset_map = None

        self.group_radius = None

    def infer_dims(self, r_data, angle_data, group_data):
        rs = np.sort(np.unique(r_data))
        angles = np.sort(np.unique(angle_data))
        groups = np.sort(np.unique(group_data)) if group_data is not None else None

        self.set_dims(rs, angles, groups)        

    def set_dims(self, rs, angles, groups):
        self.angles = angles
        self.xangles = self.angle_transform(angles)
        self.angle_map = dict(zip(self.angles, self.xangles))

        self.rs = rs
        num_rs = len(rs)

        # map r value to radius
        if self.inner_radius is None:
            self.inner_radius = 1.0 / ( 2 * num_rs )

        hdr = ( 1.0 - self.inner_radius ) / num_rs / 2.0
        self.radii = np.linspace(self.inner_radius + hdr, 
                                 1.0 - hdr, 
                                 num_rs)

        self.r_radius_map = dict(zip(rs, self.radii))
        self.group_radius = hdr * self.group_scale
        self.groups = groups if groups is not None else [ np.nan ]
        num_groups = len(self.groups)

        # map group to group offset
        if num_groups == 1:
            self.group_offsets = [ [ 0, 0 ] ]
        else:
            offset_radius = self.group_radius * self.circle_scale

            self.group_offsets = polar_linspace(offset_radius/np.sqrt(2), 
                                                -45, -45-360, num_groups)

            self.group_radius = offset_radius * 0.5

        self.group_offset_map = dict(zip(self.groups, self.group_offsets))


    def show_axes(self, angles=None, radii=None, closed=False, color=None):
        ax = plt.gca()

        if self.angles is None:
            raise Exception("dimensions not set!")

        if color is None:
            color = DEFAULT_AXIS_COLOR

        if angles is None:
            angles = self.xangles

        if radii is None:
            radii = self.radii
        
        lines = angle_lines(angles, radii[0], radii[-1])
        lines.set_zorder(1)
        lines.set_edgecolors(color)
        ax.add_collection(lines)

        if closed:
            collection = radial_circles(radii)
        else:
            collection = radial_arcs(radii, min(angles), max(angles))
            
        collection.set_facecolors((0,0,0,0.0))
        collection.set_edgecolors(color)
        collection.set_zorder(1)
        ax.add_collection(collection)


    def show_angle_labels(self, angles=None, labels=None, color=None, offset=.05, fontdict=None):
        if angles is None:
            angles = self.xangles

        if labels is None:
            labels = self.angles.astype(int)
        
        if color is None:
            color = DEFAULT_LABEL_COLOR

        add_angle_labels(plt.gca(), angles, labels, 1.0, offset=offset, color=color, fontdict=fontdict)


    def show_group_labels(self, groups=None, color=None, fontdict=None):
        ax = plt.gca()

        if groups is None:
            groups = self.groups

        if color is None:
            color = DEFAULT_LABEL_COLOR

        r = self.inner_radius*.5
        angle = 90.0
        
        x = r * np.cos(angle)
        y = r * np.sin(angle)

        for group in groups:
            off = self.group_offset_map[group]

            xfm = mxfms.Affine2D().translate(r+off[0]*2.0,off[1]*2.0).rotate(self.angle_transform(angle))
            p = xfm.transform_point([0,0])

            ax.text(p[0], p[1],
                    group, color=color, 
                    horizontalalignment='center', 
                    verticalalignment='center',
                    fontdict=fontdict)

        start_theta = self.angle_transform(angle+20)
        end_theta = self.angle_transform(angle-20)

        ax.add_patch(mpatches.Arc((0,0), 2*r, 2*r,
                                  theta1=start_theta*180.0/np.pi,
                                  theta2=end_theta*180.0/np.pi,
                                  color=color))

        ax.add_collection(LineCollection([[[0, .7*r], [0, 1.3*r]]], color=color))



        
    def show_r_labels(self, radii=None, labels=None, color=None, offset=.1, fontdict=None):
        ax = plt.gca()

        if radii is None:
            radii = self.radii

        if labels is None:
            labels = self.rs
            
        if color is None:
            color = DEFAULT_LABEL_COLOR

        if labels is None:
            labels = self.rs

        line_th = self.xangles[0]
        line_x = radii * np.cos(line_th)
        line_y = radii * np.sin(line_th)
        for i,(x,y) in enumerate(zip(line_x,line_y)):
            ax.text(x, y-offset,
                    labels[i], color=color, 
                    horizontalalignment='center', 
                    verticalalignment='center',
                    fontdict=fontdict)

    def plot(self, 
             r_data,
             angle_data,
             group_data=None,
             data=None, 
             cmap=DEFAULT_COLOR_MAP,
             clim=None,
             rmap=None,
             rlim=None,
             axis_color=None,
             label_color=None):

        ax = plt.gca()

        if data is None:
            data = np.ones(len(r_data))

        clim = self._clim(clim, data)

        if rmap is None:
            rnorm = np.vectorize(lambda x: 1.0)
        else:
            if rlim is None:
                rlim = clim
            norm = mcolors.Normalize(clim[0], clim[1])
            rnorm = np.vectorize(lambda x: rmap(norm(x)))

        if self.angles is None:
            self.infer_dims(r_data, angle_data, group_data)

        num_groups = len(self.groups)
        num_rs = len(self.rs)
        num_angles = len(self.angles)

        df = pd.DataFrame({ 'group': group_data, 
                            'angle': angle_data, 
                            'r': r_data })

        # compute circle radius
        trials_per_group = float(len(df)) / num_groups / num_rs / num_angles
        rings = rings_in_hex_pack(trials_per_group)
        circle_radius = self.group_radius / (2*rings - 1) * self.circle_scale

        gb = df.groupby(['group', 'angle', 'r'])

        for (group, angle, r), trials in gb.groups.iteritems():
            responses = np.sort(data[trials])[::-1]

            circles = spiral_trials_polar(self.r_radius_map[r],
                                          self.angle_map[angle],
                                          rnorm(responses) * circle_radius,
                                          offset=self.group_offset_map[group])

            circles.set_transform(circles.get_transform() + ax.transData)
            circles.set_array(responses)
            circles.set_cmap(cmap)
            circles.set_clim(clim)
            circles.set_zorder(2)
            circles.set_linewidths(0)

            ax.add_collection(circles)

        self.finalize()

    @staticmethod
    def for_static_gratings():
        return FanPlotter(angle_start=180, 
                          plot_scale=0.9, 
                          circle_scale=2.0,
                          group_scale=0.4,
                          plot_center=[0,.45],
                          inner_radius=.2)

    @staticmethod
    def for_drifting_gratings():
        return FanPlotter()


    


    



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from scipy.ndimage.filters import gaussian_filter
import numpy as np
import scipy.interpolate as spinterp
from .tools import dict_generator
from allensdk.api.cache import memoize
import os
import warnings
from skimage.measure import block_reduce

def upsample_image_to_degrees(img):

    upsample = 74.4/img.shape[0]
    x = np.arange(img.shape[0])
    y = np.arange(img.shape[1])

    g = spinterp.interp2d(y, x, img, kind='linear')
    ZZ_on = g(np.arange(0, img.shape[1], 1. / upsample), np.arange(0, img.shape[0], 1. / upsample))

    return ZZ_on

def convolve(img, sigma=4):
    '''
    2D Gaussian convolution
    '''

    if img.sum() == 0:
        return img

    img_pad = np.zeros((3 * img.shape[0], 3 * img.shape[1]))
    img_pad[img.shape[0]:2 * img.shape[0], img.shape[1]:2 * img.shape[1]] = img

    x = np.arange(3 * img.shape[0])
    y = np.arange(3 * img.shape[1])
    g = spinterp.interp2d(y, x, img_pad, kind='linear')

    if img.shape[0] == 16:
        upsample = 4
        offset = -(1 - .625)
    elif img.shape[0] == 8:
        upsample = 8
        offset = -(1 - .5625)
    else:
        raise NotImplementedError
    ZZ_on = g(offset + np.arange(0, img.shape[1] * 3, 1. / upsample), offset + np.arange(0, img.shape[0] * 3, 1. / upsample))
    ZZ_on_f = gaussian_filter(ZZ_on, float(sigma), mode='constant')

    z_on_new = block_reduce(ZZ_on_f, (upsample, upsample))
    z_on_new = z_on_new / z_on_new.sum() * img.sum()
    z_on_new = z_on_new[img.shape[0]:2 * img.shape[0], img.shape[1]:2 * img.shape[1]]

    return z_on_new



@memoize
def get_A(data, stimulus):

    stimulus_table = data.get_stimulus_table(stimulus)
    stimulus_template = data.get_stimulus_template(stimulus)[stimulus_table['frame'].values, :,:]

    number_of_pixels = stimulus_template.shape[1]*stimulus_template.shape[2]

    A = np.zeros((2*number_of_pixels, stimulus_template.shape[0]))
    for fi in range(stimulus_template.shape[0]):
        A[:number_of_pixels, fi] = (stimulus_template[fi,:,:].flatten() > 127).astype(float)
        A[number_of_pixels:, fi] = (stimulus_template[fi, :, :].flatten() < 127).astype(float)

    return A

@memoize
def get_A_blur(data, stimulus):

    stimulus_table = data.get_stimulus_table(stimulus)
    stimulus_template = data.get_stimulus_template(stimulus)[stimulus_table['frame'].values, :, :]

    A = get_A(data, stimulus).copy()


    number_of_pixels = A.shape[0] / 2
    for fi in range(A.shape[1]):
        A[:number_of_pixels,fi] = convolve(A[:number_of_pixels, fi].reshape(stimulus_template.shape[1], stimulus_template.shape[2])).flatten()
        A[number_of_pixels:,fi] = convolve(A[number_of_pixels:, fi].reshape(stimulus_template.shape[1], stimulus_template.shape[2])).flatten()


    return A

def get_shuffle_matrix(data, event_vector, A, number_of_shuffles=5000, response_detection_error_std_dev=.1):

    number_of_events = event_vector.sum()
    number_of_pixels = A.shape[0] / 2
    shuffle_data = np.zeros((2*number_of_pixels, number_of_shuffles))
    for ii in range(number_of_shuffles):

        size = number_of_events + int(np.round(response_detection_error_std_dev*number_of_events*np.random.randn()))
        shuffled_event_inds = np.random.choice(range(len(event_vector)), size=size, replace=False)
        b_tmp = np.zeros(len(event_vector))
        b_tmp[shuffled_event_inds] = 1
        shuffle_data[:, ii] = A.dot(b_tmp)/float(size)

    return shuffle_data

def get_sparse_noise_epoch_mask_list(st, number_of_acquisition_frames, threshold=7):

    delta = (st.start.values[1:] - st.end.values[:-1])
    cut_inds = np.where(delta > threshold)[0] + 1

    epoch_mask_list = []

    if len(cut_inds) > 2:
        warnings.warn('more than 2 epochs cut')
        print('    ', len(delta), cut_inds)

    for ii in range(len(cut_inds)+1):

        if ii == 0:
            first_ind = st.iloc[0].start
        else:
            first_ind = st.iloc[cut_inds[ii-1]].start

        if ii == len(cut_inds):
            last_ind_inclusive = st.iloc[-1].end
        else:
            last_ind_inclusive = st.iloc[cut_inds[ii]-1].end

        epoch_mask_list.append((first_ind,last_ind_inclusive))

    return epoch_mask_list

def smooth(x,window_len=11,window='hanning', mode='valid'):
    """smooth the data using a window with requested size.

    This method is based on the convolution of a scaled window with the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the begining and end part of the output signal.

    input:
        x: the input signal
        window_len: the dimension of the smoothing window; should be an odd integer
        window: the type of window from 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'
            flat window will produce a moving average smoothing.

    output:
        the smoothed signal

    example:

    t=linspace(-2,2,0.1)
    x=sin(t)+randn(len(t))*0.1
    y=smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman, numpy.convolve
    scipy.signal.lfilter

    TODO: the window parameter could be the window itself if an array instead of a string
    NOTE: length(output) != length(input), to correct this: return y[(window_len/2-1):-(window_len/2)] instead of just y.
    """

    if x.ndim != 1:
        raise ValueError, "smooth only accepts 1 dimension arrays."

    if x.size < window_len:
        raise ValueError, "Input vector needs to be bigger than window size."


    if window_len<3:
        return x


    if not window in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
        raise ValueError, "Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'"


    s=np.r_[x[window_len-1:0:-1],x,x[-1:-window_len:-1]]
    #print(len(s))
    if window == 'flat': #moving average
        w=np.ones(window_len,'d')
    else:
        w=eval('np.'+window+'(window_len)')

    y=np.convolve(w/w.sum(),s,mode=mode)
    return y


def get_components(receptive_field_data):

    s1, s2 = receptive_field_data.shape

    candidate_pixel_list = np.where(receptive_field_data.flatten()==True)[0]
    pixel_coord_dict = dict((px, (px/s2, (px - s2 * (px/s2)), px% (s1 * s2) == px)) for px in candidate_pixel_list)

    component_list = []

    for curr_pixel in candidate_pixel_list:

        curr_x, curr_y, curr_frame = pixel_coord_dict[curr_pixel]

        component_list.append([curr_pixel])
        dist_to_component_dict = {}
        for ii, curr_component in enumerate(component_list):
            dist_to_component_dict[ii] = np.inf
            for other_pixel in curr_component:

                other_x, other_y, other_frame = pixel_coord_dict[other_pixel]

                if other_frame == curr_frame:
                    x_dist = np.abs(curr_x - other_x)
                    y_dist = np.abs(curr_y - other_y)
                    curr_dist = max(x_dist, y_dist)
                    if curr_dist <  dist_to_component_dict[ii]:
                        dist_to_component_dict[ii] = curr_dist

        # Merge all components with a distance leq 1 to current pixel
        new_component_list = []
        tmp = []
        for ii, curr_component in enumerate(component_list):
            if dist_to_component_dict[ii] <= 1:
                tmp += curr_component
            else:
                new_component_list.append(curr_component)

        new_component_list.append(tmp)
        component_list = new_component_list


    if len(component_list) == 0:
        return np.zeros((1, receptive_field_data.shape[0], receptive_field_data.shape[1])), len(component_list)
    elif len(component_list) == 1:
        return_array = np.zeros((1,receptive_field_data.shape[0], receptive_field_data.shape[1]))
    else:
        return_array = np.zeros((len(component_list), receptive_field_data.shape[0], receptive_field_data.shape[1]))

    for ii, component in enumerate(component_list):
        curr_component_mask = np.zeros_like(receptive_field_data, dtype=np.bool).flatten()
        curr_component_mask[component] = True
        return_array[ii,:,:] = curr_component_mask.reshape(receptive_field_data.shape)



    return return_array, len(component_list)

def get_attribute_dict(rf):

    attribute_dict = {}
    for x in dict_generator(rf):
        if x[-3] == 'attrs':
            if len(x[:-3]) == 0:
                key = x[-2]
            else:
                key = '/'.join(['/'.join(x[:-3]), x[-2]])
            attribute_dict[key] = x[-1]

    return attribute_dict




# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import allensdk.core.json_utilities as ju
import logging
from allensdk.config.manifest import Manifest
import pandas as pd
import six


class ManifestBuilder(object):
    df_columns = ['key', 'parent_key', 'spec', 'type', 'format']

    def __init__(self):
        self._log = logging.getLogger(__name__)
        self.path_info = []
        self.sections = {}

    def set_version(self, value):
        self.path_info.append({'type': Manifest.VERSION, 'value': value})

    def add_path(self, key, spec,
                 typename='dir',
                 parent_key=None,
                 format=None):
        entry = {
            'key': key,
            'type': typename,
            'spec': spec}

        if format is not None:
            entry['format'] = format

        if parent_key is not None:
            entry['parent_key'] = parent_key

        self.path_info.append(entry)

    def add_section(self, name, contents):
        self.sections[name] = contents

    def write_json_file(self, path, overwrite=False):
        mode = 'wb'

        if overwrite is True:
            mode = 'wb+'

        json_string = self.write_json_string()

        with open(path, mode) as f:
            try:
                f.write(json_string)   # Python 2.7
            except TypeError:
                f.write(bytes(json_string, 'utf-8'))  # Python 3

    def get_config(self):
        wrapper = {"manifest": self.path_info}
        for section in self.sections.values():
            wrapper.update(section)

        return wrapper

    def get_manifest(self):
        return Manifest(self.path_info)

    def write_json_string(self):
        config = self.get_config()
        return ju.write_string(config)

    def as_dataframe(self):
        return pd.DataFrame(self.path_info,
                            columns=ManifestBuilder.df_columns)

    def from_dataframe(self, df):
        self.path_info = {}

        for _, k, p, s, t, f in six.iteritems(df.loc[:, ManifestBuilder.df_columns]):
            self.add_path(k, s, typename=t, parent=p, format=f)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2016-2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import scipy.stats as st
import pandas as pd
import numpy as np
import h5py
from .stimulus_analysis import StimulusAnalysis
from .brain_observatory_exceptions import MissingStimulusException
from . import stimulus_info as stiminfo
from . import circle_plots as cplots

class NaturalMovie(StimulusAnalysis):
    """ Perform tuning analysis specific to natural movie stimulus.

    Parameters
    ----------
    data_set: BrainObservatoryNwbDataSet object

    movie_name: string
        one of [ stimulus_info.NATURAL_MOVIE_ONE, stimulus_info.NATURAL_MOVIE_TWO,
                 stimulus_info.NATURAL_MOVIE_THREE ]
    """

    def __init__(self, data_set, movie_name, **kwargs):
        super(NaturalMovie, self).__init__(data_set, **kwargs)
        
        self.movie_name = movie_name
        self._sweeplength = NaturalMovie._PRELOAD
        self._sweep_response = NaturalMovie._PRELOAD

    @property
    def sweeplength(self):
        if self._sweeplength is NaturalMovie._PRELOAD:
            self.populate_stimulus_table()

        return self._sweeplength

    @property
    def sweep_response(self):
        if self._sweep_response is NaturalMovie._PRELOAD:
            self._sweep_response = self.get_sweep_response()

        return self._sweep_response

    def populate_stimulus_table(self):
        stimulus_table = self.data_set.get_stimulus_table(self.movie_name)
        self._stim_table = stimulus_table[stimulus_table.frame == 0]
        self._sweeplength = \
            self.stim_table.start.iloc[1] - self.stim_table.start.iloc[0]

    def get_sweep_response(self):
        ''' Returns the dF/F response for each cell

        Returns
        -------
        Numpy array
        '''
        sweep_response = pd.DataFrame(index=self.stim_table.index.values, columns=np.array(
            range(self.numbercells)).astype(str))
        for index, row in self.stim_table.iterrows():
            start = row.start
            end = start + self.sweeplength
            for nc in range(self.numbercells):
                sweep_response[str(nc)][index] = self.dfftraces[nc, start:end]
        return sweep_response

    def get_peak(self):
        ''' Computes properties of the peak response condition for each cell.

        Returns
        -------
        Pandas data frame with the below fields. A suffix of "nm1", "nm2" or "nm3" is appended to the field name depending
        on which of three movie clips was presented.
            * peak_nm1 (frame with peak response)
            * response_variability_nm1
        '''
        peak_movie = pd.DataFrame(index=range(self.numbercells), columns=(
            'peak', 'response_reliability', 'cell_specimen_id'))
        cids = self.data_set.get_cell_specimen_ids()

        mask = np.ones((10,10))
        for i in range(10):
            for j in range(10):
                if i>=j:
                    mask[i,j] = np.NaN        
        
        for nc in range(self.numbercells):
            peak_movie.cell_specimen_id.iloc[nc] = cids[nc]
            meanresponse = self.sweep_response[str(nc)].mean()
            
#            movie_len = len(meanresponse) / 30
#            output = np.empty((movie_len, 10))
#            for tr in range(10):
#                test = self.sweep_response[str(nc)].iloc[tr]
#                for i in range(movie_len):
#                    _, p = st.ks_2samp(
#                        test[i * 30:(i + 1) * 30], test[(i + 1) * 30:(i + 2) * 30])
#                    output[i, tr] = p
#            output = np.where(output < 0.05, 1, 0)
#            ptime = np.sum(output, axis=1)
#            ptime *= 10
            peak = np.argmax(meanresponse)
#            if peak > 30:
#                peak_movie.response_reliability.iloc[
#                    nc] = ptime[(peak - 30) / 30]
#            else:
#                peak_movie.response_reliability.iloc[nc] = ptime[0]
            peak_movie.peak.iloc[nc] = peak
            
            #reliability
            corr_matrix = np.empty((10,10))
            for i in range(10):
                for j in range(10):
                    r,p = st.pearsonr(self.sweep_response[str(nc)].iloc[i], self.sweep_response[str(nc)].iloc[j])
                    corr_matrix[i,j] = r
            corr_matrix*=mask
            peak_movie.response_reliability.iloc[nc] = np.nanmean(corr_matrix)
            
        if self.movie_name == stiminfo.NATURAL_MOVIE_ONE:
            peak_movie.rename(columns={
                              'peak': 'peak_'+stiminfo.NATURAL_MOVIE_ONE_SHORT, 
                              'response_reliability': 'response_reliability_'+stiminfo.NATURAL_MOVIE_ONE_SHORT}, 
                              inplace=True)
        elif self.movie_name == stiminfo.NATURAL_MOVIE_TWO:
            peak_movie.rename(columns={
                              'peak': 'peak_'+stiminfo.NATURAL_MOVIE_TWO_SHORT, 
                              'response_reliability': 'response_reliability_'+stiminfo.NATURAL_MOVIE_TWO_SHORT},
                              inplace=True)
        elif self.movie_name == stiminfo.NATURAL_MOVIE_THREE:
            peak_movie.rename(columns={
                              'peak': 'peak_'+stiminfo.NATURAL_MOVIE_THREE_SHORT, 
                              'response_reliability': 'response_reliability_'+stiminfo.NATURAL_MOVIE_THREE_SHORT}, 
                              inplace=True)

        return peak_movie

    def open_track_plot(self, cell_specimen_id=None, cell_index=None):
        cell_index = self.row_from_cell_id(cell_specimen_id, cell_index)

        cell_rows = self.sweep_response[str(cell_index)]
        data = []
        for i in range(len(cell_rows)):
            data.append(cell_rows.iloc[i])

        data = np.vstack(data)

        tp = cplots.TrackPlotter(ring_length=360)
        tp.plot(data,
                clim=[0, data.mean() + data.std()*3])
        tp.show_arrow()

    @staticmethod 
    def from_analysis_file(data_set, analysis_file, movie_name):
        nm = NaturalMovie(data_set, movie_name)
        nm.populate_stimulus_table()

        # TODO: deal with this properly
        suffix_map = {
            stiminfo.NATURAL_MOVIE_ONE: '_'+stiminfo.NATURAL_MOVIE_ONE_SHORT,
            stiminfo.NATURAL_MOVIE_TWO: '_'+stiminfo.NATURAL_MOVIE_TWO_SHORT,
            stiminfo.NATURAL_MOVIE_THREE: '_'+stiminfo.NATURAL_MOVIE_THREE_SHORT
            }

        try:
            suffix = suffix_map[movie_name]


            nm._sweep_response = pd.read_hdf(analysis_file, "analysis/sweep_response"+suffix)
            nm._peak = pd.read_hdf(analysis_file, "analysis/peak")

            with h5py.File(analysis_file, "r") as f:
                nm._binned_dx_sp = f["analysis/binned_dx_sp"].value
                nm._binned_cells_sp = f["analysis/binned_cells_sp"].value
                nm._binned_dx_vis = f["analysis/binned_dx_vis"].value
                nm._binned_cells_vis = f["analysis/binned_cells_vis"].value
        except Exception as e:
            raise MissingStimulusException(e.args)

        return nm



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from ..biophys_sim.config import Config
from .utils import create_utils
from allensdk.core.nwb_data_set import NwbDataSet
import allensdk.ephys.extract_cell_features as extract_cell_features
from shutil import copy
import numpy
import logging
import time
import os

_runner_log = logging.getLogger('allensdk.model.biophysical.runner')


def run(description, sweeps=None):
    '''Main function for running a biophysical experiment.

    Parameters
    ----------
    description : Config
        All information needed to run the experiment.
    '''
    model_type = description.data['biophys'][0]['model_type']

    # configure NEURON
    utils = create_utils(description, model_type)
    h = utils.h

    # configure model
    manifest = description.manifest
    morphology_path = description.manifest.get_path('MORPHOLOGY')
    utils.generate_morphology(morphology_path.encode('ascii', 'ignore'))
    utils.load_cell_parameters()

    # configure stimulus and recording
    stimulus_path = description.manifest.get_path('stimulus_path')
    run_params = description.data['runs'][0]
    if sweeps is None:
        sweeps = run_params['sweeps']
    sweeps_by_type = run_params['sweeps_by_type']
    junction_potential = description.data['fitting'][0]['junction_potential']
    mV = 1.0e-3

    prepare_nwb_output(manifest.get_path('stimulus_path'),
                       manifest.get_path('output_path'))

    # run sweeps
    for sweep in sweeps:
        _runner_log.info("Running sweep: %d" % (sweep))
        utils.setup_iclamp(stimulus_path, sweep=sweep)
        _runner_log.info("Done loading sweep: %d" % (sweep))
        vec = utils.record_values()
        tstart = time.time()
        h.finitialize()
        h.run()
        tstop = time.time()
        _runner_log.info("Time: %f" % (tstop - tstart))

        # write to an NWB File
        output_data = (numpy.array(vec['v']) - junction_potential) * mV

        output_path = manifest.get_path("output_path")

        save_nwb(output_path, output_data, sweep, sweeps_by_type)


def prepare_nwb_output(nwb_stimulus_path,
                       nwb_result_path):
    '''Copy the stimulus file, zero out the recorded voltages and spike times.

    Parameters
    ----------
    nwb_stimulus_path : string
        NWB file name
    nwb_result_path : string
        NWB file name
    '''

    output_dir = os.path.dirname(nwb_result_path)
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    copy(nwb_stimulus_path, nwb_result_path)
    data_set = NwbDataSet(nwb_result_path)
    data_set.fill_sweep_responses(0.0)
    for sweep in data_set.get_sweep_numbers():
        data_set.set_spike_times(sweep, [])


def save_nwb(output_path, v, sweep, sweeps_by_type):
    '''Save a single voltage output result into an existing sweep in a NWB file.
    This is intended to overwrite a recorded trace with a simulated voltage.

    Parameters
    ----------
    output_path : string
        file name of a pre-existing NWB file.
    v : numpy array
        voltage
    sweep : integer
        which entry to overwrite in the file.
    '''
    output = NwbDataSet(output_path)
    output.set_sweep(sweep, None, v)

    sweep_by_type = {t: [sweep]
                     for t, ss in sweeps_by_type.items() if sweep in ss}
    sweep_features = extract_cell_features.extract_sweep_features(output,
                                                                  sweep_by_type)
    try:
        spikes = sweep_features[sweep]['spikes']
        spike_times = [s['threshold_t'] for s in spikes]
        output.set_spike_times(sweep, spike_times)
    except Exception as e:
        logging.info("sweep %d has no sweep features. %s" % (sweep, e.args))


def load_description(manifest_json_path):
    '''Read configuration file.

    Parameters
    ----------
    manifest_json_path : string
        File containing the experiment configuration.

    Returns
    -------
    Config
        Object with all information needed to run the experiment.
    '''
    description = Config().load(manifest_json_path)

    # fix nonstandard description sections
    fix_sections = ['passive', 'axon_morph,', 'conditions', 'fitting']
    description.fix_unary_sections(fix_sections)

    return description


if '__main__' == __name__:
    import sys

    description = load_description(sys.argv[-1])
    run(description)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2015-2016. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import os
from six import string_types

from allensdk.config.manifest_builder import ManifestBuilder
from allensdk.api.cache import Cache
from allensdk.api.queries.cell_types_api import CellTypesApi

from . import json_utilities as json_utilities
from .nwb_data_set import NwbDataSet
from . import  swc

import logging
import warnings
import pandas as pd


class CellTypesCache(Cache):
    """
    Cache class for storing and accessing data from the Cell Types Database.
    By default, this class will cache any downloaded metadata or files in
    well known locations defined in a manifest file.  This behavior can be
    disabled.

    Attributes
    ----------

    api: CellTypesApi instance
        The object used for making API queries related to the Cell Types Database

    Parameters
    ----------

    cache: boolean
        Whether the class should save results of API queries to locations specified
        in the manifest file.  Queries for files (as opposed to metadata) must have a
        file location.  If caching is disabled, those locations must be specified
        in the function call (e.g. get_ephys_data(file_name='file.nwb')).

    manifest_file: string
       File name of the manifest to be read.  Default is "cell_types_manifest.json".
    """

    # manifest keys
    CELLS_KEY = 'CELLS'
    EPHYS_FEATURES_KEY = 'EPHYS_FEATURES'
    MORPHOLOGY_FEATURES_KEY = 'MORPHOLOGY_FEATURES'
    EPHYS_DATA_KEY = 'EPHYS_DATA'
    EPHYS_SWEEPS_KEY = 'EPHYS_SWEEPS'
    RECONSTRUCTION_KEY = 'RECONSTRUCTION'
    MARKER_KEY = 'MARKER'
    MANIFEST_VERSION = None

    def __init__(self, cache=True, manifest_file='cell_types_manifest.json', base_uri=None):
        super(CellTypesCache, self).__init__(
            manifest=manifest_file, cache=cache, version=self.MANIFEST_VERSION)
        self.api = CellTypesApi(base_uri=base_uri)

    def get_cells(self, file_name=None,
                  require_morphology=False,
                  require_reconstruction=False,
                  reporter_status=None):
        """
        Download metadata for all cells in the database and optionally return a
        subset filtered by whether or not they have a morphology or reconstruction.

        Parameters
        ----------

        file_name: string
            File name to save/read the cell metadata as JSON.  If file_name is None,
            the file_name will be pulled out of the manifest.  If caching
            is disabled, no file will be saved. Default is None.

        require_morphology: boolean
            Filter out cells that have no morphological images.

        require_reconstruction: boolean
            Filter out cells that have no morphological reconstructions.

        reporter_status: list
            Filter for cells that have one or more cell reporter statuses.
        """

        file_name = self.get_cache_path(file_name, self.CELLS_KEY)

        if os.path.exists(file_name):
            cells = json_utilities.read(file_name)
        else:
            cells = self.api.list_cells(False, False)

            if self.cache:
                json_utilities.write(file_name, cells)

        if isinstance(reporter_status, string_types):
            reporter_status = [reporter_status]

        # filter the cells on the way out
        return self.api.filter_cells(cells,
                                     require_morphology,
                                     require_reconstruction,
                                     reporter_status)
        

    def get_ephys_sweeps(self, specimen_id, file_name=None):
        """
        Download sweep metadata for a single cell specimen.

        Parameters
        ----------

        specimen_id: int
             ID of a cell.
        """

        file_name = self.get_cache_path(
            file_name, self.EPHYS_SWEEPS_KEY, specimen_id)

        sweeps = self.api.get_ephys_sweeps(specimen_id, 
                                           strategy='lazy', 
                                           path=file_name, 
                                           **Cache.cache_json())

        return sweeps

    def get_ephys_features(self, dataframe=False, file_name=None):
        """
        Download electrophysiology features for all cells in the database.

        Parameters
        ----------

        file_name: string
            File name to save/read the ephys features metadata as CSV.
            If file_name is None, the file_name will be pulled out of the
            manifest.  If caching is disabled, no file will be saved.
            Default is None.

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return
            a list of dictionaries.
        """
        file_name = self.get_cache_path(file_name, self.EPHYS_FEATURES_KEY)

        if self.cache:
            if dataframe:
                warnings.warn("dataframe argument is deprecated.")
                args = Cache.cache_csv_dataframe()
            else:
                args = Cache.cache_csv_json()
            args['strategy'] = 'lazy'
        else:
            args = Cache.nocache_json()

        features_df = self.api.get_ephys_features(path=file_name,
                                                  **args)

        return features_df


    def get_morphology_features(self, dataframe=False, file_name=None):
        """
        Download morphology features for all cells with reconstructions in the database.

        Parameters
        ----------

        file_name: string
            File name to save/read the ephys features metadata as CSV.
            If file_name is None, the file_name will be pulled out of the
            manifest.  If caching is disabled, no file will be saved.
            Default is None.

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return
            a list of dictionaries.
        """

        file_name = self.get_cache_path(
            file_name, self.MORPHOLOGY_FEATURES_KEY)

        if self.cache:
            if dataframe:
                warnings.warn("dataframe argument is deprecated.")
                args = Cache.cache_csv_dataframe()
            else:
                args = Cache.cache_csv_json()
        else:
            args = Cache.nocache_json()

        args['strategy'] = 'lazy'
        args['path'] = file_name

        return self.api.get_morphology_features(**args)


    def get_all_features(self, dataframe=False, require_reconstruction=True):
        """
        Download morphology and electrophysiology features for all cells and merge them
        into a single table.

        Parameters
        ----------

        dataframe: boolean
            Return the output as a Pandas DataFrame.  If False, return
            a list of dictionaries.

        require_reconstruction: boolean
            Only return ephys and morphology features for cells that have
            reconstructions. Default True.
        """
        
        ephys_features = pd.DataFrame(self.get_ephys_features())
        morphology_features = pd.DataFrame(self.get_morphology_features())

        how = 'inner' if require_reconstruction else 'outer'

        all_features = ephys_features.merge(morphology_features,
                                            how=how,
                                            on='specimen_id')

        if dataframe:
            warnings.warn("dataframe argument is deprecated.")
            return all_features
        else:
            return all_features.to_dict('records')

    def get_ephys_data(self, specimen_id, file_name=None):
        """
        Download electrophysiology traces for a single cell in the database.

        Parameters
        ----------

        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the ephys features metadata as CSV.
            If file_name is None, the file_name will be pulled out of the
            manifest.  If caching is disabled, no file will be saved.
            Default is None.

        Returns
        -------
        NwbDataSet
            A class instance with helper methods for retrieving stimulus
            and response traces out of an NWB file.
        """

        file_name = self.get_cache_path(
            file_name, self.EPHYS_DATA_KEY, specimen_id)

        self.api.save_ephys_data(specimen_id, file_name, strategy='lazy')

        return NwbDataSet(file_name)

    def get_reconstruction(self, specimen_id, file_name=None):
        """
        Download and open a reconstruction for a single cell in the database.

        Parameters
        ----------

        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the reconstruction SWC.
            If file_name is None, the file_name will be pulled out of the
            manifest.  If caching is disabled, no file will be saved.
            Default is None.

        Returns
        -------
        Morphology
             A class instance with methods for accessing morphology compartments.
        """

        file_name = self.get_cache_path(
            file_name, self.RECONSTRUCTION_KEY, specimen_id)

        if file_name is None:
            raise Exception(
                "Please enable caching (CellTypes.cache = True) or specify a save_file_name.")

        if not os.path.exists(file_name):
            self.api.save_reconstruction(specimen_id, file_name)

        return swc.read_swc(file_name)

    def get_reconstruction_markers(self, specimen_id, file_name=None):
        """
        Download and open a reconstruction marker file for a single cell in the database.

        Parameters
        ----------

        specimen_id: int
            The ID of a cell specimen to download.

        file_name: string
            File name to save/read the reconstruction marker.
            If file_name is None, the file_name will be pulled out of the
            manifest.  If caching is disabled, no file will be saved.
            Default is None.

        Returns
        -------
        Morphology
             A class instance with methods for accessing morphology compartments.
        """

        file_name = self.get_cache_path(
            file_name, self.MARKER_KEY, specimen_id)

        if file_name is None:
            raise Exception(
                "Please enable caching (CellTypes.cache = True) or specify a save_file_name.")

        if not os.path.exists(file_name):
            try:
                self.api.save_reconstruction_markers(specimen_id, file_name)
            except LookupError as e:
                logging.warning(e.args)
                return []

        return swc.read_marker_file(file_name)

    def build_manifest(self, file_name):
        """
        Construct a manifest for this Cache class and save it in a file.

        Parameters
        ----------

        file_name: string
            File location to save the manifest.

        """

        mb = ManifestBuilder()
        mb.set_version(self.MANIFEST_VERSION)
        mb.add_path('BASEDIR', '.')
        mb.add_path(self.CELLS_KEY, 'cells.json',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_DATA_KEY, 'specimen_%d/ephys.nwb',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_FEATURES_KEY, 'ephys_features.csv',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.MORPHOLOGY_FEATURES_KEY, 'morphology_features.csv',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.RECONSTRUCTION_KEY, 'specimen_%d/reconstruction.swc',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.MARKER_KEY, 'specimen_%d/reconstruction.marker',
                    typename='file', parent_key='BASEDIR')
        mb.add_path(self.EPHYS_SWEEPS_KEY, 'specimen_%d/ephys_sweeps.json',
                    typename='file', parent_key='BASEDIR')

        mb.write_json_file(file_name)


class ReporterStatus:
    """
    Valid strings for filtering by cell reporter status.
    """

    POSITIVE = 'cre reporter positive'
    NEGATIVE = 'cre reporter negative'
    NA = 'not applicable'
    INDETERMINATE = 'cre reporter indeterminate'



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
from ..api import Api


class TreeSearchApi(Api):
    '''

    See `Searching a Specimen or Structure Tree <http://help.brain-map.org/display/api/Image-to-Image+Synchronization>`_
    for additional documentation.
    '''

    def __init__(self, base_uri=None):
        super(TreeSearchApi, self).__init__(base_uri)

    def get_tree(self,
                 kind,
                 db_id,
                 ancestors=None,
                 descendants=None):
        '''Fetch meta data for the specified structure or specimen.

        Parameters
        ----------
        kind : string
            'Structure' or 'Specimen'
        db_id : integer
            The id of the structure or specimen to search.
        ancestors : boolean, optional
            whether to include ancestors in the response (defaults to False)
        descendants : boolean, optional
            whether to include descendants in the response (defaults to False)

        Returns
        -------
        dict
            parsed json response data
        '''
        params = []
        url_params = ''

        if ancestors is True:
            params.append('ancestors=true')
        elif ancestors is False:
            params.append('ancestors=false')

        if descendants is True:
            params.append('descendants=true')
        elif descendants is False:
            params.append('descendants=false')

        if len(params) > 0:
            url_params = '?' + '&'.join(params)
        else:
            url_params = ''

        url = ''.join([self.tree_search_endpoint,
                       '/',
                       kind,
                       '/',
                       str(db_id),
                       '.json',
                       url_params])

        return self.json_msg_query(url)



# Allen Institute Software License - This software license is the 2-clause BSD
# license plus a third clause that prohibits redistribution for commercial
# purposes without further permission.
#
# Copyright 2017. Allen Institute. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Redistributions for commercial purposes are not permitted without the
# Allen Institute's written permission.
# For purposes of this license, commercial purposes is the incorporation of the
# Allen Institute's software into anything for which you will charge fees or
# other compensation. Contact terms@alleninstitute.org for commercial licensing
# opportunities.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
import pytest
from mock import patch, mock_open, MagicMock
from allensdk.core.brain_observatory_cache import (BrainObservatoryCache, 
                                                   _find_container_tags,
                                                   _merge_transgenic_lines,
                                                   _find_specimen_cre_line,
                                                   _find_specimen_reporter_line)

try:
    import __builtin__ as builtins  # @UnresolvedImport
except:
    import builtins  # @UnresolvedImport


CACHE_MANIFEST = """
{
  "manifest": [
    {
      "type": "dir",
      "spec": ".",
      "key": "BASEDIR"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "experiment_containers.json",
      "key": "EXPERIMENT_CONTAINERS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_experiments.json",
      "key": "EXPERIMENTS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "ophys_experiment_data/%d.nwb",
      "key": "EXPERIMENT_DATA"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "cell_specimens.json",
      "key": "CELL_SPECIMENS"
    },
    {
      "parent_key": "BASEDIR",
      "type": "file",
      "spec": "stimulus_mappings.json",
      "key": "STIMULUS_MAPPINGS"
    }
  ]
}
"""


@pytest.fixture
def brain_observatory_cache():
    boc = None

    try:
        manifest_data = bytes(CACHE_MANIFEST, 'UTF-8')  # Python 3
    except:
        manifest_data = bytes(CACHE_MANIFEST)  # Python 2.7

    with patch('os.path.exists',
               return_value=True):
        with patch(builtins.__name__ + ".open",
                   mock_open(read_data=manifest_data)):
            # Download a list of all targeted areas
            boc = BrainObservatoryCache(manifest_file='boc/manifest.json',
                                        base_uri='http://api.brain-map.org')

    boc.api.json_msg_query = MagicMock(name='json_msg_query')

    return boc

def test_get_all_targeted_structures(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                brain_observatory_cache.get_all_targeted_structures()

        brain_observatory_cache.api.json_msg_query.assert_called_once_with(
            "http://api.brain-map.org/api/v2/data/query.json?q="
            "model::ExperimentContainer,rma::include,"
            "ophys_experiments,isi_experiment,"
            "specimen(donor(conditions,age,transgenic_lines)),"
            "targeted_structure,"
            "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_experiment_containers(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download experiment containers for VISp experiments
                visp_ecs = brain_observatory_cache.get_experiment_containers(
                    targeted_structures=['VISp'])

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ExperimentContainer,rma::include,"
        "ophys_experiments,isi_experiment,"
        "specimen(donor(conditions,age,transgenic_lines)),targeted_structure,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_all_cre_lines(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all cre lines
                tls = brain_observatory_cache.get_all_cre_lines()

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ExperimentContainer,rma::include,"
        "ophys_experiments,isi_experiment,"
        "specimen(donor(conditions,age,transgenic_lines)),targeted_structure,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_ophys_experiments(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache.get_ophys_experiments()

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::OphysExperiment,rma::include,experiment_container,"
        "well_known_files(well_known_file_type),targeted_structure,"
        "specimen(donor(age,transgenic_lines)),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_all_session_types(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache.get_all_session_types()

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::OphysExperiment,rma::include,experiment_container,"
        "well_known_files(well_known_file_type),targeted_structure,"
        "specimen(donor(age,transgenic_lines)),"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


def test_get_stimulus_mappings(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            with patch('allensdk.core.json_utilities.read',
                       MagicMock(name='read_json')):
                # Download a list of all transgenic driver lines
                tls = brain_observatory_cache._get_stimulus_mappings()

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q="
        "model::ApiCamStimulusMapping,"
        "rma::options[num_rows$eq'all'][count$eqfalse]")


@pytest.mark.skipif(True, reason="need to develop mocks")
def test_get_cell_specimens(brain_observatory_cache):
    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.core.json_utilities.write',
                   MagicMock(name='write_json')):
            # Download a list of all transgenic driver lines
            tls = brain_observatory_cache.get_cell_specimens()

    brain_observatory_cache.api.json_msg_query.assert_called_once_with(
        "http://api.brain-map.org/api/v2/data/query.json?q=")


def test_build_manifest():
    try:
        manifest_data = bytes(CACHE_MANIFEST, 'UTF-8')  # Python 3
    except:
        manifest_data = bytes(CACHE_MANIFEST)  # Python 2.7

    with patch('os.path.exists') as m:
        m.return_value = False

        with patch('allensdk.config.manifest.Manifest.safe_mkdir') as mkdir:
            with patch('allensdk.config.manifest_builder.'
                       'ManifestBuilder.write_json_file',
                       MagicMock(name='write_json_file')) as mock_write_json:
                with patch(builtins.__name__ + ".open",
                           mock_open(read_data=manifest_data)):
                    brain_observatory_cache = BrainObservatoryCache(
                        manifest_file='boc/manifest.json',
                        base_uri='http://api.brain-map.org')
                    mkdir.assert_called_once_with('boc')
                    mock_write_json.assert_called_once_with(
                        'boc/manifest.json')


def test_string_argument_errors(brain_observatory_cache):
    boc = brain_observatory_cache

    with pytest.raises(TypeError):
        boc.get_experiment_containers(targeted_structures='str')

    with pytest.raises(TypeError):
        boc.get_experiment_containers(cre_lines='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(targeted_structures='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(cre_lines='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(stimuli='str')

    with pytest.raises(TypeError):
        boc.get_ophys_experiments(session_types='str')

def test_find_container_tags():
    # no conditions no tags
    c = { "specimen": { "donor": { "conditions": [] } } }
    tags = _find_container_tags(c)
    assert len(tags) == 0

    # tissue tags are ignored
    c = { "specimen": { "donor": { "conditions": [ { "name": "tissuecyte" } ] } } }
    tags = _find_container_tags(c)
    assert len(tags) == 0

    # no conditions is okay
    c = { "specimen": { "donor": { } } }
    tags = _find_container_tags(c)
    assert len(tags) == 0

    # everything else goes through
    c = { "specimen": { "donor": { "conditions": [ { "name": "fish" } ] } } }
    tags = _find_container_tags(c)
    assert len(tags) == 1

def test_merge_transgenic_lines():
    # None is allowed and should be ignored
    t1 = [ "a", "b", "c" ]
    t2 = None
    tm = _merge_transgenic_lines(t1,t2)
    assert sorted(tm) == [ "a", "b", "c"]

    # otherwise it's just a merge
    t1 = [ "a", "b", "c" ]
    t2 = [ "c", "d" ]
    tm = _merge_transgenic_lines(t1,t2)
    assert sorted(tm) == [ "a", "b", "c", "d"]

    # one list is fine
    t1 = [ "a", "b", "c" ]
    tm = _merge_transgenic_lines(t1)
    assert sorted(tm) == [ "a", "b", "c" ]

def test_find_specimen_cre_line():
    # None if no TLs
    s = { "donor": { "transgenic_lines": [ ] } }
    cre = _find_specimen_cre_line(s)
    assert cre is None

    # None if no 'Cre'
    s = { "donor": { "transgenic_lines": [ { "transgenic_line_type_name": "driver", "name": "banana" } ] } }
    cre = _find_specimen_cre_line(s)
    assert cre is None

    # None if no 'Cre'
    s = { "donor": { "transgenic_lines": [ { "transgenic_line_type_name": "driver", "name": "bananaCre" } ] } }
    cre = _find_specimen_cre_line(s)
    assert cre == "bananaCre"

    # None if no 'driver'
    s = { "donor": { "transgenic_lines": [ { "transgenic_line_type_name": "reporter", "name": "bananaCre" } ] } }
    cre = _find_specimen_cre_line(s)
    assert cre == None

def test_find_specimen_reporter_line():
    # None if no TLs
    s = { "donor": { "transgenic_lines": [ ] } }
    cre = _find_specimen_reporter_line(s)
    assert cre is None

    s = { "donor": { "transgenic_lines": [ { "transgenic_line_type_name": "reporter", "name": "banana" } ] } }
    cre = _find_specimen_reporter_line(s)
    assert cre == "banana"

    # None if no "reporter"
    s = { "donor": { "transgenic_lines": [ { "transgenic_line_type_name": "driver", "name": "bananaCre" } ] } }
    cre = _find_specimen_reporter_line(s)
    assert cre is None


    
